#!/usr/bin/env python3
# coding: utf-8
import os.path
from os import path
from collections import defaultdict
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Alphabet import IUPAC
from Bio import SeqIO
import pysam
import sys

"""
Command line arguments
"""
blatMapping = sys.argv[1] # Blat result file from microMappingPipeline.py
summaryMapping = sys.argv[2] # Summary of the blat mapping from microMappingPipeline.py
outputDir = sys.argv[3] # output directory ...
genomeQuery = sys.argv[4] # fasta file of the v1 genome (must be fai indexed)
genomeTarget = sys.argv[5] # fasta file of the v2 genome (must be fai indexed)
annotQuery = sys.argv[6] # bed file of new coords of markers on target reference genome

"""
Global var
"""
geneAnnot=defaultdict()
geneMapping=defaultdict()
newCoord=defaultdict()
outputGff=outputDir+'/targetAnnot.gff3'


def main():

	"""
	check for input files and output directory
	"""
	sys.stdout.write(' Check input files:')
	checkFile(file=blatMapping)
	checkFile(file=summaryMapping)
	checkFile(file=genomeQuery)
	checkFile(file=genomeTarget)
	checkFile(file=annotQuery)

	"""
	open for reading the query and target reference file
	This will be used to extract the dna sequences before and after recalculation to ensure that the sequences are the same
	and that the calculation is correct
	"""
	sys.stdout.write(' Open FASTA files for sequence extraction\n')
	queryFasta = pysam.FastaFile(genomeQuery)
	targetFasta = pysam.FastaFile(genomeTarget)

	"""
	Save blat mapping info into dict
	"""
	geneMapping=saveBlatResults(summary=summaryMapping, blat=blatMapping)

	"""
	Save Annotation infirmation from GFF file
	"""
	geneAnnot=saveAnnotFromGFF(gff=annotQuery)

	"""
	For each gene, with full match (with or without missmatches)
	"""
	for geneID in geneAnnot.keys():
		print(" Dealing with gene {}\n".format(geneID))

		blat=''
		annot=geneAnnot[geneID]

		if geneID in geneMapping.keys():
			blat=geneMapping[geneID]
		else:
			sys.stderr.srite(" ERROR cannot find gene {} in blat data\n".format(geneID))
			continue

		numFeatures = len(annot)

		# some logs
		print(" Annotation features:")
		print(annot)
		print(" Found {} features for gene {}\n".format(numFeatures,geneID))
		print(" Blat mapping:")
		print(blat)

		'''
		Recalc the new coord
		'''
		newCoord = recalcCoord(
			strand=blat['hitStrandOnTarget'],
			queryChrom=blat['queryChrom'],
			queryStart=blat['queryStart'],
			queryStop=blat['queryStop'],
			targetChrom=blat['targetChrom'],
			targetStart=blat['targetStart'],
			targetStop=blat['targetStop'],
			hitStart=blat['hitStartOnTarget'],
			hitStop=blat['hitStopOnTarget'],
			annot=annot
			)
		input()


def recalcCoord(strand,queryChrom,queryStart,queryStop,targetChrom,targetStart,targetStop,hitStart,hitStop,annot):
	newFeatureCoord=defaultdict()

	if strand == '+':
		print(' Hit on strand + of the target')
		print(' target chrom {} starting from {} and ending at {}'.format(targetChrom,targetStart,targetStop))

		for feature in annot:
			feature=feature.rstrip('\n').split("\t")
			print(' recalc coord for feature {}'.format(feature))

			newStart=targetStart + (feature[3] - queryStart)
			newStop=targetStart + (feature[4] - queryStart)

	elif strand == '-':
		print(' Hit on strand - of the target')
		print(' target chrom {} starting from {} and ending at {}'.format(targetChrom,targetStart,targetStop))

		for feature in annot:
			feature=feature.rstrip('\n').split("\t")
			print(' recalc coord for feature {}'.format(feature))

			newStart=targetStop + (feature[3] - queryStart)
			newStop=targetStop + (feature[4] - queryStart)

			print(' New coords on ')

	else:
		sys.stderr.write(' Cannot parse strand of blat hit on target ')



def saveAnnotFromGFF(gff):
	sys.stdout.write(" 2. saving Query Annotation from file {} into Dict \n".format(gff))
	sys.stdout.write("===========================================================================\n")
	dataDict=defaultdict()
	geneID=''
	with open(gff) as gffRecords:
		for line in gffRecords.readlines():
			if line.startswith('#'):
				continue

			# transform gff line into array
			(chrom,source,featureType,start,stop,score,strand,frame,info)=line.rstrip('\n').split('\t')

			# save the last column with info tags
			gffInfoTable=info.split(';')

			# and transform it into a dict
			gffInfoDict=defaultdict()
			for infoPair in gffInfoTable:
				(key,value)=infoPair.split('=')
				gffInfoDict[key] = value

			if featureType == 'gene':
				geneID=gffInfoDict['ID']
				dataDict[geneID] = [line.rstrip('\n')]

			else:
				dataDict[geneID].append(line.rstrip('\n'))

	numGenes=len(dataDict.keys())
	sys.stdout.write(" Found {} genes in GFF file {}\n".format(numGenes,gff))
	return dataDict

def saveBlatResults(summary, blat):
	sys.stdout.write(" 1. saving blat results {} and summary {} into Dict \n".format(blat,summary))
	sys.stdout.write("===========================================================================\n")
	dataDict=defaultdict()

	# SUMMARY DATA
	with open(summary) as summaryData:
		for line in summaryData.readlines():
			#sys.stdout.write(' current record is {}'.format(line))

			(geneName,markerPair,mappingStatus,queryChrom,markerStatus,markerPair2,geneName2,geneQueryStart,geneQueryEnd,geneQueryStrand,leftMakerId,leftMakerQueryLoc,leftMarkerTargetLoc,rightMarkerId,rightMakerQueryLoc,rightMarkerTargetLoc) = line.rstrip('\n').split("\t")
			
			dataDict[geneName] = {
				'geneId': geneName,
				'gffRecords': 'NA',
				'gffSubfeatures': [],
				'targetStart':'NA',
				'targetStop':'NA',
				'targetChrom':'NA',
				'hitStrandOnTarget':'NA',
				'hitStartOnTarget':'NA',
				'hitStopOnTarget':'NA',
				'hitStrandOnTarget':'NA',
				'queryStart':geneQueryStart,
				'queryStop':geneQueryEnd,
				'queryChrom':queryChrom,
				'queryStrand':geneQueryStrand,
				'mappingStatus':mappingStatus
			}
	numGenesInDict=len(dataDict.keys())

	# BLAT MAPPING DATA
	with open(blat) as blatData:
		for line in blatData.readlines():

			(geneName,markerPair,mappingStatus,mappingStatus2,blatCoverage,blatIndelBases,blatMissmatches,hitStartOnTarget,hitStopOnTarget,hitStrand,targetLength,targetName)=line.rstrip('\n').split('\t')
			
			if geneName in dataDict.keys():
				
				if mappingStatus == 'unmapped':
					sys.stdout.write(" gene {} is not mapped: no target information to recover.\n".format(geneName))
					continue
				else:
					(prefix,targetChrom,targetStart,targetStop)=targetName.split('_')
					dataDict[geneName]['targetStart'] = targetStart
					dataDict[geneName]['targetStop'] = targetStop
					dataDict[geneName]['targetChrom'] = targetChrom
					dataDict[geneName]['hitStrandOnTarget'] = hitStrand
					dataDict[geneName]['hitStartOnTarget'] = hitStartOnTarget
					dataDict[geneName]['hitStopOnTarget'] = hitStopOnTarget

			else:
				sys.stderr.write(" ERROR: gene {} found in blat results but not in summary data! \n".format(geneName))
				sys.exit()

	sys.stdout.write(" Number of records in Gene Dict based on summaryData: {}\n".format(numGenesInDict))
	return dataDict
	
def checkFile (file):
	if os.path.isfile(file):
		sys.stdout.write(" File %s found\n" % str(file))
	else:
		sys.stdout.write(" ERROR: Cannot find file %s \n" % str(file))
		sys.exit()

def checkDir(directory):
	if os.path.isdir(directory):
		sys.stdout.write(" Directory %s found\n" % str(directory))
		return 1
	else:
		sys.stdout.write(" Cannot find directory %s \n" % str(directory))
		return 0


if __name__== "__main__":
	if ( checkDir(directory=outputDir) ):
		sys.stdout.write( " 0. Output directory already exists, no need to create it\n")

	else:
		sys.stdout.write( " 0. Creating output directory %s" % str(outputDir))
		os.mkdir( outputDir, 0o750 )

	outputGffFH=open(outputGff, 'w')

	main()